"""
SPEats - a meal planning and recipe management app

Modules:
- main.py: main program
- db.py: database models and setup
- config.py: paths to directories and configuration settings
- data_handler.py: reads data from Excel and populates the database

To run the program, execute main.py in the terminal.

=======================

Author: Tyson Johnson
Date: 2024-06-25
Version: 1.0
"""

import argparse
from cds_302_project.db import Database
from cds_302_project.data_handler import DataHandler
from cds_302_project.config import database_file, entry_file

def main(drop_tables=False):
    # Define database
    sp_eats_db = Database(database_file)
    #sp_eats_db.create_db()

    with sp_eats_db.session_scope() as session:
        if drop_tables:
            # Drop tables if they exist (only in development or testing)
            sp_eats_db.drop_tables()

        # Create tables if they don't exist
        sp_eats_db.create_tables()

        # Create objects and populate the database
        data_handler = DataHandler(entry_file, session)
        data_handler.create_objects()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='SPEats - Meal Planning and Recipe Management App')
    parser.add_argument('--drop-tables', action='store_true', help='Drop tables before creating them')
    args = parser.parse_args()

    main(drop_tables=args.drop_tables)



"""
This file contains classes that represent the tables in the database. 
Each class represents a table in the database and has attributes that correspond to the columns in the table.
The classes also have methods that can be used to interact with the database.
"""

from sqlalchemy import Column, Integer, String, Numeric, ForeignKey, create_engine, MetaData
from sqlalchemy.orm import sessionmaker, declarative_base, relationship
from contextlib import contextmanager
from sqlalchemy_utils import database_exists, create_database

Base = declarative_base()

class Database:
    def __init__(self, address):
        self.engine = create_engine('sqlite:///' + address.as_posix())
        self.Session = sessionmaker(bind=self.engine)
        self.metadata = MetaData()

    def create_db(self):
        if not database_exists(self.engine.url):
            create_database(self.engine.url)

    def create_tables(self):
        Base.metadata.create_all(self.engine)

    def drop_tables(self):
        list_of_table_names = ['dish', 'ingredient', 'dish_ingredient', 'dietary_restriction', 'ingredient_dietary_restriction', 'meal_type', 'dish_meal_type'][::-1]
        for table_name in list_of_table_names:
            Base.metadata.tables[table_name].drop(self.engine)

    @contextmanager
    def session_scope(self):
        session = self.Session()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

# Define database models
class Ingredient(Base):
    __tablename__ = 'ingredient'
    ingredient_id = Column(Integer, primary_key=True)
    name = Column(String(64))
    calories = Column(Numeric(10,2))
    fat = Column(Numeric(10,2))
    carbs = Column(Numeric(10,2))
    protein = Column(Numeric(10,2))
    alcohol = Column(Numeric(10,2))
    fiber = Column(Numeric(10,2))

class Dish(Base):
    __tablename__ = 'dish'
    dish_id = Column(Integer, primary_key=True)
    name = Column(String(64))
    calories = Column(Numeric(10,2))
    fat = Column(Numeric(10,2))
    carbs = Column(Numeric(10,2))
    protein = Column(Numeric(10,2))
    alcohol = Column(Numeric(10,2))
    fiber = Column(Numeric(10,2))
    ingredients = relationship("DishIngredient", back_populates="dish")

class DishIngredient(Base):
    __tablename__ = 'dish_ingredient'
    dish_id = Column(Integer, ForeignKey('dish.dish_id'), primary_key=True)
    ingredient_id = Column(Integer, ForeignKey('ingredient.ingredient_id'), primary_key=True)
    quantity = Column(Numeric(10,2))
    calories = Column(Numeric(10,2))
    fat = Column(Numeric(10,2))
    carbs = Column(Numeric(10,2))
    protein = Column(Numeric(10,2))
    alcohol = Column(Numeric(10,2))
    fiber = Column(Numeric(10,2))
    dish = relationship("Dish", back_populates="ingredients")
    ingredient = relationship("Ingredient")

class DietaryRestriction(Base):
    __tablename__ = 'dietary_restriction'
    restriction_id = Column(Integer, primary_key=True)
    name = Column(String(64))

class IngredientDietaryRestriction(Base):
    __tablename__ = 'ingredient_dietary_restriction'
    ingredient_id = Column(Integer, ForeignKey('ingredient.ingredient_id'), primary_key=True)
    restriction_id = Column(Integer, ForeignKey('dietary_restriction.restriction_id'), primary_key=True)

class MealType(Base):
    __tablename__ = 'meal_type'
    type_id = Column(Integer, primary_key=True)
    name = Column(String(64))

class DishMealType(Base):
    __tablename__ = 'dish_meal_type'
    dish_id = Column(Integer, ForeignKey('dish.dish_id'), primary_key=True)
    type_id = Column(Integer, ForeignKey('meal_type.type_id'), primary_key=True)



"""
This file contains the configuration settings for the project.
"""

from pathlib import Path
import os
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv()

# Define directories
working_dir = Path.cwd() / "cds_302_project"
data_dir = working_dir / "data"
output_dir = working_dir / "output"
logs_dir = output_dir / "logs"
entry_file = data_dir / "food_data.xlsx"

# This is for publishing online, requires environment variables to be set
# from sqlalchemy import engine
# database_file = engine.URL.create(
#         host=os.getenv('env_HOST'),
#         database=os.getenv('env_DATABASE'),
#         username=os.getenv('env_USER'),
#         password=os.getenv('env_PASS'), 
#         drivername=os.getenv('env_DRIVER')
#     )

database_file = data_dir / "cds302.db"

# Ensure all directories exist
data_dir.mkdir(parents=True, exist_ok=True)
output_dir.mkdir(parents=True, exist_ok=True)
logs_dir.mkdir(parents=True, exist_ok=True)

# Logging configuration
log_file = logs_dir / 'data_handler.log'
log_level = os.getenv('LOG_LEVEL', 'INFO').upper()


import pandas as pd
import logging
from cds_302_project.db import Ingredient, Dish, DishIngredient, DietaryRestriction, IngredientDietaryRestriction, MealType, DishMealType
from cds_302_project.config import log_file, log_level

# Set up logging
logging.basicConfig(filename=log_file, level=log_level, format='%(asctime)s %(levelname)s:%(message)s')

class DataHandler:
    def __init__(self, entry_file: str, session):
        self.entry_file = entry_file
        self.session = session
        self.data = {
            "ingredient": self.read_data("ingredient", ["ingredient_id", "name", "calories", "fat", "carbs", "protein", "alcohol", "fiber"]),
            "dish": self.read_data("dish", ["dish_id", "name", "calories", "fat", "carbs", "protein", "alcohol", "fiber"]),
            "dish_ingredient": self.read_data("dish_ingredient", ["dish_id", "ingredient_id", "quantity", "calories", "fat", "carbs", "protein", "alcohol", "fiber"]),
            "dietary_restriction": self.read_data("dietary_restriction", ["restriction_id", "name"]),
            "ingredient_dietary_restriction": self.read_data("ingredient_dietary_restriction", ["ingredient_id", "restriction_id"]),
            "meal_type": self.read_data("meal_type", ["type_id", "name"]),
            "dish_meal_type": self.read_data("dish_meal_type", ["dish_id", "type_id"])
        }

    def read_data(self, sheet_name_var: str, usecols_var: list) -> list:
        dtype_var = {col: int if "id" in col else str for col in usecols_var}
        logging.info(f"Reading data from sheet: {sheet_name_var}")
        return pd.read_excel(self.entry_file, sheet_name=sheet_name_var, usecols=usecols_var, dtype=dtype_var).to_dict('records')

    def create_objects(self):
        def add_record(session, model, record):
            if not session.query(model).get(record['ingredient_id']):
                session.add(model(**record))
            else:
                logging.info(f"Skipping {model.__name__} with ID {record['ingredient_id']} as it already exists.")

        for data in self.data["ingredient"]:
            add_record(self.session, Ingredient, data)
        self.session.commit()

        for data in self.data["dish"]:
            if not self.session.query(Dish).get(data['dish_id']):
                self.session.add(Dish(**data))
            else:
                logging.info(f"Skipping Dish with ID {data['dish_id']} as it already exists.")
        self.session.commit()

        for data in self.data["dietary_restriction"]:
            if not self.session.query(DietaryRestriction).get(data['restriction_id']):
                self.session.add(DietaryRestriction(**data))
            else:
                logging.info(f"Skipping DietaryRestriction with ID {data['restriction_id']} as it already exists.")
        self.session.commit()

        for data in self.data["meal_type"]:
            if not self.session.query(MealType).get(data['type_id']):
                self.session.add(MealType(**data))
            else:
                logging.info(f"Skipping MealType with ID {data['type_id']} as it already exists.")
        self.session.commit()

        for data in self.data["dish_ingredient"]:
            if not self.session.query(DishIngredient).get((data['dish_id'], data['ingredient_id'])):
                self.session.add(DishIngredient(**data))
            else:
                logging.info(f"Skipping DishIngredient with Dish ID {data['dish_id']} and Ingredient ID {data['ingredient_id']} as it already exists.")
        self.session.commit()

        for data in self.data["ingredient_dietary_restriction"]:
            if not self.session.query(IngredientDietaryRestriction).get((data['ingredient_id'], data['restriction_id'])):
                self.session.add(IngredientDietaryRestriction(**data))
            else:
                logging.info(f"Skipping IngredientDietaryRestriction with Ingredient ID {data['ingredient_id']} and Restriction ID {data['restriction_id']} as it already exists.")
        self.session.commit()

        for data in self.data["dish_meal_type"]:
            if not self.session.query(DishMealType).get((data['dish_id'], data['type_id'])):
                self.session.add(DishMealType(**data))
            else:
                logging.info(f"Skipping DishMealType with Dish ID {data['dish_id']} and Type ID {data['type_id']} as it already exists.")
        self.session.commit()
